"""
mypythontools
=============

Some tools/functions/snippets used across projects.

Usually used from IDE. Root path is infered and things like docs generation on pre-commit
githook, building application with pyinstaller or deploying to Pypi is matter of calling one function.

Many projects - one codebase.

If you are not sure whether structure of app that will work with this code, there is python starter repo
on `https://github.com/Malachov/my-python-starter`

Modules:
--------

build
-----
Build your app with pyinstaller just with calling one function `build_app`.
Check function doctrings for how to do it.

See module help for more informations.

githooks
--------

Some functions runned every each git action (usually before commit).

Can derive run pytest tests, `README.md` from `__init__.py` or generate rst files necessary for sphinx docs generator.
All with one line

Check module docstrings for how to use it.

deploy
------

Deploy app on Pypi.


pyvueel
-------
Common functions for Python / Vue / Eel project.

Run `mypythontools.pyvueel.help_starter_pack_vue_app()` for tutorial how to create such an app.


misc
----
Miscellaneous
Set up root path if not cwd for example.
"""

from . import githooks
from . import build
from . import misc
from . import pyvueel

__version__ = "0.0.5"

__author__ = "Daniel Malachov"
__license__ = "MIT"
__email__ = "malachovd@seznam.cz"

__all__ = ['githooks', 'build', 'deploy', 'misc', 'pyvueel']



"""
Allow to deploy project.
Possible destinations: PYPI.
"""
import subprocess
import os

from . import misc


def deploy_to_pypi(setup_path=None):
    """Publish python library to Pypi. Username and password are set
    with env vars `TWINE_USERNAME` and `TWINE_PASSWORD`.

    Args:
        setup_path((str, pathlib.Path)): Function suppose, that there is a setup.py in cwd. If not, pass path to setup.py.
        username
        password
    """
    setup_path = misc.root_path if not setup_path else setup_path

    subprocess.run(['python', 'setup.py', 'sdist', 'bdist_wheel'], cwd=misc.root_path, shell=True, check=True)
    subprocess.run(['twine', 'upload', '-u', os.environ['TWINE_USERNAME'], '-p', os.environ['TWINE_PASSWORD'], 'dist/*'], cwd=misc.root_path, shell=True, check=True)



"""
This module is to be used in githooks (some code automatically executed before commit for example).

How to
------

Create folder git_hooks with git hook file - for prec commit name must be `pre-commit` (with no extension)
Hooks in git folder are gitignored by default (and hooks is not visible on first sight).

Then add hook to git settings - run in terminal (last arg is path (created folder))

```shell
git config core.hooksPath git_hooks
```

In created folder on first two lines copy this

```python
#!/usr/bin/env python
# -*- coding: UTF-8 -*-
```

Then just import any function from here and call with desired params. E.g.

Examples:
    >>> mypythontools.githooks.run_tests()
    >>> mypythontools.githooks.generate_readme_from_init('mylibrary')
    >>> mypythontools.githooks.sphinx_docs_regenerate('mylibrary')

That will generate readme from __init__.py and call sphinx-apidoc and create rst files ind doc source folder.
"""

import subprocess
import inspect
from pathlib import Path
import importlib
import pytest

from . import misc


def run_tests(test_path=None):
    """Run tests and if not passing, raise an error.

    Args:
        test_path ((str, pathlib.Path)), optional): Usually autodetected (if cwd / tests). Defaults to None.

    Raises:
        Exception: If any test fail, it will raise exception (git hook do not continue...).
    """

    if not test_path:
        test_path = misc.root_path / 'tests'

    pytested = pytest.main(["-x", test_path.as_posix()])

    if pytested == 1:
        raise Exception("Pytest failed")


def sphinx_docs_regenerate(project_name, root_path=None, app_path=None, docs_path=None, build_locally=0, git_add=True):
    """This will generate all rst files necessary for sphinx documentation generation.
    It automatically delete removed and renamed files.

    Function suppose sphinx build and source in separate folders...

    Args:
        root_path ((str, Path), optional): Where docs folder is. Usually infered automatically. Defaults to None.
        app_path ((str, Path), optional): Where application scripts are. Usually infered automatically. Defaults to None.
        docs_path ((str, Path), optional): Where source folder is. Usually infered automatically. Defaults to None.
        project_name (str): Name of your project (used in sphinx-apidoc).
        build_locally (bool, optional): If true, build build folder with html files locally. Defaults to 0.
        git_add (bool, optional): Whether to add generated files to stage. False mostly for testing reasons.
    Note:
        Function suppose structure of docs like

        -- docs
        -- -- source
        -- -- -- conf.py
        -- -- make.bat

        If you are issuing error, try set project root path with `set_root`
    """

    if not importlib.util.find_spec('sphinx'):
        raise ImportError("Sphinx library is necessary for docs generation. Install via `pip install sphinx`")

    if not root_path:
        root_path = misc.root_path
    if not docs_path:
        docs_path = root_path / 'docs'

    docs_source_path = root_path / 'docs' / 'source'

    if not app_path:
        app_path = root_path / project_name

    for p in Path(docs_source_path).iterdir():
        if p.name not in ['conf.py', 'index.rst', '_static', '_templates']:
            p.unlink()

    if build_locally:
        subprocess.run(['make', 'html'], shell=True, cwd=docs_path, check=True)

    subprocess.run(['sphinx-apidoc', '-f', '-e', '-o', 'source', app_path.as_posix()], shell=True, cwd=docs_path, check=True)

    if git_add:
        subprocess.run(['git', 'add', 'docs'], shell=True, cwd=root_path, check=True)


def generate_readme_from_init(project_name, git_add=True):
    """Because i had very similar things in main __init__.py and in readme. It was to maintain news in code.
    For better simplicity i prefer write docs once and then generate. One code, two use cases.

    Why __init__? - Because in IDE on mouseover developers can see help.
    Why README.md? - Good for github.com

    Args:
        project_name (str): Module name (module must be imported).
        git_add (bool, optional): Whether to add generated files to stage. False mostly for testing reasons.
    """

    ## Generate README.md from __init__.py
    my_module = importlib.import_module(project_name)
    docstrings = inspect.getdoc(my_module)

    root_path = misc.root_path

    with open(root_path / 'README.md', 'w') as file:
        file.write(docstrings)

    if git_add:
        subprocess.run(['git', 'add', 'README.md'], shell=True, cwd=root_path, check=True)



"""
Only internal helping module for other modules. Not supposed to be used by users.
"""
from pathlib import Path
import sys


# Root is usually current working directory, if not, use `set_root` function.
root_path = Path.cwd()


def set_root(set_root_path=None):
    """Root folder is inferred automatically if call is from git_hooks folder or from root (cwd).
    If more projects opened in IDE, root project path can be configured here.

    Args:
        root_path ((str, pathlib.Path)): Path to project root.
    """
    if set_root_path:
        root_path = Path(set_root_path)

    if not root_path.as_posix() in sys.path:
        sys.path.insert(0, root_path.as_posix())


def find_file(file, folder=None):
    """Look on files in folder (cwd() by default) and find file with it's folder.

    Args:
        file (str): Name with extension e.g. "app.py".

    Returns:
        Path: Path of file.
    """
    # TODO exclude node_modules or restrict nested level
    file = Path(file)

    folder = root_path if not folder else folder

    for i in root_path.glob('**/*'):
        if i.name == file.name:
            return i



"""
Common functions for Python / Vue / Eel project.

Run `mypythontools.pyvueel.help_starter_pack_vue_app()` for tutorial how to create such an app.

"""

import os
import sys
from pathlib import Path
import warnings
import pandas as pd
import numpy as np

import mylogging

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', module='eel', category=ResourceWarning)

    import eel


def run_gui(multiprocessing=False, log_file_path=None):
    """Function that init and run `eel` project.
    It will autosetup chrome mode (if installed chrome or chromium, open separate window with
    no url bar, no bookmarks etc...) if not chrome installed, it open microsoft Edge (by default on windows).

    # TODO document more
    # In devel mode, app is connected on live vue server.

    Args:
        multiprocessing (bool, optional): If using multiprocessing in some library, set up to True. Defaults to False.
        log_file_path ((str, Path, None)), optional): If not exist, it will create, if exist, it will append,
        if None, log to relative log.log and only if in production mode.
    """

    try:
        devel = False if os.environ.get('MY_PYTHON_VUE_ENVIRONMENT') == 'production' else True

        if log_file_path:
            log_file = log_file_path
        else:
            log_file = 'log.log' if not devel else None

        mylogging.config.TO_FILE = log_file

        if getattr(sys, 'frozen', False):
            gui_path = Path(sys._MEIPASS) / 'gui'
        else:
            gui_path = Path(__file__).resolve().parent / 'gui'

        if devel:
            directory = gui_path / 'src'
            app = None
            page = {'port': 8080}
            port = 8686

            def close_callback(page, sockets):
                pass

        else:
            directory = gui_path / 'web_builded'
            close_callback = None
            app = 'chrome'
            page = 'index.html'
            port = 0

        eel.init(directory.as_posix(), ['.vue', '.js', '.html'])

        # try:
        # if devel:

        # If server is not running, it's started automatically
        # import psutil
        # import subprocess

        # import signal
        # already_run = 8080 in [i.laddr.port for i in psutil.net_connections()]

        # if not already_run:

        #     subprocess.Popen(f"cd '{pystore.gui_path.as_posix()}' && npm run serve", stdout=subprocess.PIPE,
        #                      shell=True)
        #     print("\nVue starting, reload page after loaded! \n")
        #     import webbrowser
        #     webbrowser.open('http://localhost:8080/', new=2)

        if multiprocessing:
            multiprocessing.freeze_support()

        mylogging.info("Py side started")

        eel.start(page, mode=app, cmdline_args=['--disable-features=TranslateUI'], close_callback=close_callback, host='localhost', port=port, disable_cache=True),

    except OSError:
        eel.start(page, mode='edge', host='localhost', close_callback=close_callback, port=port, disable_cache=True),


    except Exception:
        mylogging.traceback("Py side terminated...")


def help_starter_pack_vue_app():
    """
    Tutorial how to build app with python, Vue and eel.
    Print help on build Vue part with CLI (many options, rather not copying files from other project).
    Then will help with basic structre of components, then show copypasters to other places.

    Structure

    - myproject
        - gui
            - generated with Vue CLI
        - app.py

    ############
    ### app.py
    ###########

    from mypythontools import pyvueel

    # Expose python functions to Js with decorator
    @eel.expose
        def load_data(settings):
            # You can return dict - will be object in js
            # You can return list - will be an array in js

            return {'Hello': 1}

    # End of file
    if __name__ == '__main__':
        pyvueel.run_gui()

    #########
    ### gui
    ########

    Generate gui folder with Vue CLI

    ```console
    npm install -g @vue/cli
    vue create gui
    ```

    Goto folder and

    ```console
    vue add vuex
    vue add router
    ```

    #############
    ### main.js
    ############

    if (ProcessingInstruction.env.NODE_ENV !== 'production') {
      window.eel.set_host("ws://localhost:8686")
    }

    // You can expose function to be callable from python. Import and then
    // window.eel.expose(mutate, 'mutate')

    ##########
    ### .env
    #########

    create empty files .env.development and add `VUE_APP_EEL=http://localhost:8686/eel.js`

    create empty .env.production and add `VUE_APP_EEL=eel.js`

    #################
    ### index.html
    ###############

    In public folder

    ```
    <script type="text/javascript" src="<%= VUE_APP_EEL %>"></script>
    ```

    #################
    ### Tips, trips
    ################

    # VS Code plugins for developing
    - npm
    - vetur
    - Vue VSCode Snippets
    - vuetify-vscode
    """

    print(help_starter_pack_vue_app.__doc__)


def json_to_py(json):
    """Take json / json from JS side and eval it from strings.
    If string to string, if float to float, if object then to dict.

    When to use? - If sending object as parameter in function.

    Args:
        json (dict): Object from JS

    Returns:
        dict: Python dictionary with correct types.
    """

    evaluated = {}
    for i, j in json.items():
        try:
            evaluated[i] = eval(j)
        except Exception:
            evaluated[i] = j
    return evaluated


def to_vue_plotly(data, names=None):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly understand.

    Args:
        data ((np.array, pd.DataFrame)): Plotted data.
        names (list, optional): If using array, you can define names. Defaults to None.

    Returns:
        dict: Data in form for plotting in frontend.
    """
    if isinstance(data, (np.ndarray, np.generic)):
        data = pd.DataFrame(data, columns=names)

    numeric_data = data.select_dtypes(include='number').round(decimals=3)
    numeric_data = numeric_data.where(np.isfinite(numeric_data), None)
    # numeric_data = add_none_to_gaps(numeric_data)

    return {'x_axis': numeric_data.index.to_list(), 'y_axis': numeric_data.values.T.tolist(), 'names': numeric_data.columns.values.tolist()}


def to_table(df):
    """Takes data (dataframe or numpy array) and transforms it to form, that vue-plotly understand.

    Args:
        df (pd.DataFrame): Data in table form

    Returns:
        dict: Data in form for create table.
    """
    data = df.copy()
    data = data.round(decimals=3)

    # Numpy nan cannot be send to json - replace with None
    data = data.where(~data.isin([np.nan, np.inf, -np.inf]), None)

    headers = [{'text': i, 'value': i, 'sortable': True} for i in data.columns]

    return {'table': data.to_dict('records'), 'headers': headers}



"""
This module build the app via pyinstaller.
It help to build applications build with eel.

There is one main function `build_app`. Check it's help for how to use
it (should be very simple).

Note:
    Build pyinstaller on your pc, otherwise antivirus can check the file for a while on first run.
    Download from github, cd to bootloader and

    ```
    python ./waf all
    ```

    Back to pyinstaller folder and python `setup.py`
"""

import subprocess
import shutil
from pathlib import Path
from . import misc


def build_app(
        main_file='app.py', preset=None,
        root_path=None, web_path=None,
        build_web=None, remove_last_build=False,
        console=True, debug=False, icon=False, hidden_imports=[],
        ignored_packages=[], datas=[], name=None, env_vars={},
        cleanit=True):
    """One script to build .exe app from source code. it use pyinstaller, which must be localy installed.
    Build pyinstaller bootloader manually to avoid antivirus problems.
    This script automatically generate .spec file, build node web files and add environment variables during build.

    This script suppose some structure of the app. You can use python app starter on the same repository.

    You can use presets (or create your own.) - then you don't need to set other except the main_file.

    Args:
        main_file (str, optional): Main file path or name with extension. Main file is found automatically
            and don't have to be in root. Defaults to 'app.py'.
        root_path ((Path, str), optional): Root where build and dist folders are (as well as docs, travis.yaml are). If None,
            cwd (current working directory) infered. Defaults to None.
        web_path ((Path, str), optional): Folder with index.html
        preset (str, optional): Edit other params for specific use cases (append to hidden_imports, datas etc.)
            Options ['eel'].
        build_web (bool, optional): If application contain package.json in folder 'gui', build it (if using eel). Defaults to None.
        remove_last_build (bool, optional): If some problems, it is possible to delete build and dist folders. Defaults to False.
        console (bool, optional): Before app run terminal window appears (good for debugging). Defaults to False.
        debug (bool, optional): If no console, then dialog window with traceback appears. Defaults to False.
        icon ((Path, str), optional): Path to .ico file (!no png!). Defaults to None.
        hidden_imports (list, optional): If app is not working, it can be because some library was not builded. Add such
            libraries into this list. Defaults to [].
        ignored_packages (list, optional): Libraries take space even if not necessary. Defaults to [].
        datas (list, optional): Add static files to build. Example: [('my_source_path, 'destination_path')].
        env_vars (dict, optional): Add some env vars during build. Mostly to tell main script that it's production (ne development) mode. Defaults to {}.
        name (str, optional): If name of app is different than main py file. Defaults to None.
        cleanit (bool, optional): Remove spec file and var env py hook. Defaults to True.
    """

    # Try to recognize the structure of app
    root_path = root_path if root_path else misc.root_path

    build_path = root_path / 'build'

    if not build_path.exists():
        build_path.mkdir(parents=True, exist_ok=True)

    # May be just name - not absolute
    main_file_path = Path(main_file)

    if not main_file_path.exists():

        # Iter paths and find the one
        main_file_path = misc.find_file(main_file_path)

        if not main_file_path.exists():
            raise KeyError("Main file not found, not infered and must be configured in params...")

    main_file_path = main_file_path.resolve()

    if not name:
        name = main_file_path.stem

    main_folder_path = main_file_path.parent

    if preset == 'eel':
        if not web_path:
            for i in root_path.glob('**/*'):
                if i.name == 'index.html':
                    if i.parent != 'public':
                        web_path = i
                        break

        else:
            web_path = Path(web_path)

        if not web_path.exists():
            raise KeyError("Build web assets not found, not infered and must be configured in params...")

        if build_web is None:
            build_web = True

        hidden_imports = [*hidden_imports, 'bottle_websocket']

        # TODO
        # Find by main.js
        datas = tuple([*datas, ((main_folder_path / 'gui' / 'web_builded').as_posix(), 'gui/web_builded')])
        ignored_packages = [*ignored_packages, 'tensorflow', 'keras', 'notebook', 'pytest', 'pyzmq', 'zmq', 'sqlalchemy', 'sphinx', 'PyQt5', 'PIL', 'matplotlib', 'qt5', 'PyQt5', 'qt4', 'pillow']
        env_vars = {**env_vars, 'MY_PYTHON_VUE_ENVIRONMENT': 'production'}

    generated_warning = """
#########################
### File is generated ###
#########################

# Do not edit this file, edit build_script"""

    if env_vars:
        env_vars_template = f"""
{generated_warning}

import os
for i, j in {env_vars}.items():
    os.environ[i] = j
"""

        env_path = (build_path / 'env_vars.py')

        with open(env_path, 'w') as env_vars_py:
            env_vars_py.write(env_vars_template)
        runtime_hooks = [env_path.as_posix()]
    else:
        runtime_hooks = None

    spec_template = f"""
{generated_warning}

import sys
from pathlib import Path
import os

sys.setrecursionlimit(5000)
block_cipher = None

a = Analysis(['{main_file_path.as_posix()}'],
            pathex=['{main_folder_path.as_posix()}'],
            binaries=[],
            datas={datas},
            hiddenimports={hidden_imports},
            hookspath=[],
            runtime_hooks={runtime_hooks},
            excludes={ignored_packages},
            win_no_prefer_redirects=False,
            win_private_assemblies=False,
            cipher=block_cipher,
            noarchive=False)
pyz = PYZ(a.pure, a.zipped_data,
            cipher=block_cipher)
exe = EXE(pyz,
        a.scripts,
        [],
        exclude_binaries=True,
        name='{name}',
        debug={debug},
        bootloader_ignore_signals=False,
        strip=False,
        upx=True,
        console={console},
        icon={f'{str(icon)}' if icon else None})
coll = COLLECT(exe,
            a.binaries,
            a.zipfiles,
            a.datas,
            strip=False,
            upx=True,
            upx_exclude=[],
            name='{name}')
"""

    spec_path = build_path / 'app.spec'

    with open(spec_path, 'w') as spec_file:
        spec_file.write(spec_template)

    if remove_last_build:
        try:
            shutil.rmtree('build', ignore_errors=True)
            shutil.rmtree('dist', ignore_errors=True)
        except Exception:
            pass

    # Build JS to static asset
    if build_web:
        gui_path = misc.find_file('main.js')
        subprocess.run(['npm', 'run', 'build'], shell=True, check=True, cwd=gui_path)

    # Build py to exe
    subprocess.run(['pyinstaller', '-y', spec_path.as_posix()], shell=True, check=True, cwd=main_folder_path)

    if cleanit:
        try:
            spec_path.unlink()
            env_path.unlink()
        except Exception:
            pass



from setuptools import setup, find_packages
import pkg_resources
import mypythontools

version = mypythontools.__version__

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as f:
    myreqs = [str(requirement) for requirement in pkg_resources.parse_requirements(f)]

setup(
    name='mypythontools',
    version=version,
    url='https://github.com/Malachov/mypythontools',
    license='mit',
    author='Daniel Malachov',
    author_email='malachovd@seznam.cz',
    install_requires=myreqs,
    description='Some tools/functions/snippets used across projects.ions.',
    long_description_content_type='text/markdown',
    long_description=readme,
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Intended Audience :: Developers',
        'Intended Audience :: Education'],
    extras_require={
    }
)
